/*     Saving floppy disks as disk files. Version 2.01.
 *     (C) Fedon Kadifeli 1989-1994.
 *     Compiler: Microsoft C. Version 8.00.
 */

#include  <stdio.h>
#include  <stdlib.h>
#include  <conio.h>	/* MS-CL */

#define M_READ  (0x25)	/* absolute disk read int. no.  */
#define M_WRITE (0x26)	/* absolute disk write int. no. */
#define SECTOR  ((size_t)512) /* sector size                  */
#define IOS     ((size_t)48)  /* memory buffer size (sectors) */
#define BPS     (1024/SECTOR) /* 1 Kbyte blocks               */
#define MAX_FN  (66)	/* longest file name            */

typedef unsigned char  BYTE;  /* 8 bits */

BYTE  buf1 [IOS*SECTOR], buf2 [IOS*SECTOR];

/* comparison */
int equal (register BYTE * a, register BYTE * b,
  register size_t l)
{
  while (l--)
    if (*a++ != *b++)
      return 0;	/* do not match */
  return 1;	/* match        */
}  /* equal */

/* absolute disk input/output */
size_t aDisk (BYTE intNo, BYTE drvNo, size_t secCnt,
  size_t secNo, BYTE * trAddr)
{
  size_t retVal;
  __asm {
      push    di	; save
      push    si	;  registers
      push    bp
      mov     ah,intNo	; int. no. (M_READ or M_WRITE)
      mov     al,drvNo	; floppy disk drive number
      sub     al,'A'	; 'A' -> 0, 'B' -> 1
      mov     cx,secCnt	; no. of sectors to process
      mov     dx,secNo	; beginning sector number
      mov     bx,trAddr	; transfer address
      cmp     ah,M_READ
      jne     else	; if M_WRITE
      int     M_READ
      jmp     SHORT cont
else: int     M_WRITE
cont: pushf
      pop     ax	; get flags
      and     ax,1	; carry flag (1 : error)
      popf	; restore flags
      pop     bp	; restore
      pop     si	;  registers
      pop     di
      mov     retVal,ax
  }
  return !retVal;	/* 0 : error; 1 : OK */
}  /* aDisk */

/* ask menu options */
int askMenu (void)
{
  int  opt;
  static char * options [] = {
    "\n",
    " 1. Save\tFloppy disk --> Disk file",
    " 2. Load\tFloppy disk <-- Disk file",
    " 3. Compare\tFloppy disk <-> Disk file" };

  printf("\n\n\t%s\n\t%s\n\t%s\n\nOption:",
    options[1], options[2], options[3]);
  if ((opt = _getch()) == '1' || opt == 'S' || opt == 's')
    opt = 1;
  else if (opt == '2' || opt == 'L' || opt == 'l')
    opt = 2;
  else if (opt == '3' || opt == 'C' || opt == 'c')
    opt = 3;
  else
    opt = 0;
  puts(options[opt]);
  return opt;
}  /* askMenu */

/* ask floppy disk drive to be used */
int askDrive (void)
{
  int  drv;

  printf("Floppy disk drive (A/B): ");
  if ((drv = toupper(_getch())) != 'A' && drv != 'B')
    drv = 0;
  printf(drv ? "%c:\n" : "\n", drv);
  return drv;
}  /* askDrive */

/* ask disk file name to be used */
char * askFile (void)
{
  static char  fn [MAX_FN+3] = { MAX_FN+1 };

  printf("Disk file name: ");
  _cgets(fn);
  putchar('\n');
  return fn[2] ? fn+2 : NULL;
}  /* askFile */

int main (void)
{
  int    option, drive;
  char * fn;
  FILE * fp;
  int    b;
  size_t i, r1, r2;

  printf("\tSaving Floppy Disks As Disk Files."
    " Version 2.01.\n\t(C) Fedon Kadifeli 1989-1994.\n");
  while ((option = askMenu()) != 0)
    while ((drive = askDrive()) != 0)
      while ((fn = askFile()) != NULL) {
        i = 0;
        r1 = 1;
        b = 0;
        /* 1st option : Floppy disk --> Disk file */
        if (option == 1) {
          if ((fp = fopen(fn,"rb")) != NULL) {
            fclose(fp);
            printf("File '%s' already present!\n", fn);
            continue;
          }
          if ((fp = fopen(fn,"wb")) == NULL) {
            printf("Cannot open file '%s'!\n", fn);
            continue;
          }
          while (aDisk(M_READ,(BYTE)drive,IOS,i,buf1) &&
            (r1=fwrite((void *)buf1,SECTOR,IOS,fp)==IOS)!=0)
              printf("\r%d", (i+=IOS)/BPS);
          printf(r1 ? (i?" blocks transferred to file '%s'.\n"
                        :".  Floppy disk cannot be read!\a\n")
                    : ".  Write error on file '%s'!\a\n", fn);
        } else  {
        /* 2nd or 3rd option */
          if ((fp = fopen(fn,"rb")) == NULL) {
            printf("File '%s' cannot be opened!\n", fn);
            continue;
          }
          /* 2nd option : Floppy disk <-- Disk file */
          if (option == 2) {
            while (fread((void *)buf1,SECTOR,IOS,fp)==IOS &&
              (r1=aDisk(M_WRITE,(BYTE)drive,IOS,i,buf1))!=0)
                printf("\r%d", (i+=IOS)/BPS);
            printf(r1 ? " blocks transferred to floppy disk.\n"
                      : ".  Write error on floppy disk!\a\n");
          }  else {
          /* 3rd option : Floppy disk <-> Disk file */
            while ((r1=aDisk(M_READ,(BYTE)drive,IOS,i,buf1)) &
              (r2=fread((void *)buf2,SECTOR,IOS,fp)==IOS) &&
              (b =equal(buf1,buf2,IOS*SECTOR))!=0)
                printf("\r%d", (i+=IOS)/BPS);
            printf(
              !b ? ".  Compare error!\a\n" :
              r1 ? ".  File shorter than floppy disk!\a\n" :
              r2 ? ".  Floppy disk shorter than file!\a\n" :
                   " blocks compare OK.\n");
          }
        }  /* if */
        fclose(fp);
      }  /* while */
  printf("\nOperation complete...\n");
  return 0;
}  /* main */
